# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Changing field 'ExecutionRecord.schedule_id'
        db.alter_column('periodically_executionrecord', 'schedule_id', self.gf('django.db.models.fields.CharField')(max_length=32))


    def backwards(self, orm):
        
        # Changing field 'ExecutionRecord.schedule_id'
        db.alter_column('periodically_executionrecord', 'schedule_id', self.gf('django.db.models.fields.BigIntegerField')())


    models = {
        'periodically.executionrecord': {
            'Meta': {'object_name': 'ExecutionRecord'},
            'completed_successfully': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'end_time': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_fake': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'schedule_id': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            'scheduled_time': ('django.db.models.fields.DateTimeField', [], {}),
            'start_time': ('django.db.models.fields.DateTimeField', [], {}),
            'task_id': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        }
    }

    complete_apps = ['periodically']



from __future__ import print_function
from periodically.decorators import *

@every(minutes=1)
@hourly()
@hourly() # This repetition should have no effect.
def task1():
    print('RUNNING exampleapp.task1')


# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding model 'ExecutionRecord'
        db.create_table('periodically_executionrecord', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('task_id', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('schedule_id', self.gf('django.db.models.fields.BigIntegerField')()),
            ('scheduled_time', self.gf('django.db.models.fields.DateTimeField')()),
            ('start_time', self.gf('django.db.models.fields.DateTimeField')()),
            ('end_time', self.gf('django.db.models.fields.DateTimeField')(null=True, blank=True)),
            ('completed_successfully', self.gf('django.db.models.fields.BooleanField')(default=False)),
            ('is_fake', self.gf('django.db.models.fields.BooleanField')(default=False)),
        ))
        db.send_create_signal('periodically', ['ExecutionRecord'])


    def backwards(self, orm):
        
        # Deleting model 'ExecutionRecord'
        db.delete_table('periodically_executionrecord')


    models = {
        'periodically.executionrecord': {
            'Meta': {'object_name': 'ExecutionRecord'},
            'completed_successfully': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'end_time': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_fake': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'schedule_id': ('django.db.models.fields.BigIntegerField', [], {}),
            'scheduled_time': ('django.db.models.fields.DateTimeField', [], {}),
            'start_time': ('django.db.models.fields.DateTimeField', [], {}),
            'task_id': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        }
    }

    complete_apps = ['periodically']



from django.utils import importlib, timezone
from . import settings
from .models import ExecutionRecord


class InvalidBackendAliasError(Exception):
    def __init__(self, backend):
        Exception.__init__(self, "There is no scheduler backend with alias '%s'" % backend)


class InvalidBackendError(Exception):
    def __init__(self, backend):
        Exception.__init__(self, 'Could not find backend %s' % backend)


def get_scheduled_time(task, schedule, now=None):
    previous_record = ExecutionRecord.objects.get_most_recent(task, schedule)
    if not now:
        now = timezone.now()
    previous_scheduled_time = schedule.time_before(now)
    if not previous_record:
        scheduled_time = previous_scheduled_time
    elif previous_record.scheduled_time < previous_scheduled_time:
        scheduled_time = previous_scheduled_time
    else:
        scheduled_time = schedule.time_after(now)
    return scheduled_time    


def get_scheduler_backend_class(backend_alias=None):
    """
    Accepts a scheduler alias and returns the corresponding backend class.
    """
    if backend_alias is None:
        backend_alias = 'default'

    try:
        class_path = settings.SCHEDULERS[backend_alias].get('backend')
    except KeyError:
        raise InvalidBackendAliasError(backend_alias)

    mod_path, cls_name = class_path.rsplit('.', 1)    
    try:
        mod = importlib.import_module(mod_path)
        backend_class = getattr(mod, cls_name)
    except (AttributeError, ImportError):
        raise InvalidBackendError(class_path)
    
    return backend_class


def get_scheduler_backends_in_groups(groups):
    backend_aliases = set()
    for group in groups:
        for alias, info in settings.SCHEDULERS.items():
            if group in info.get('groups', []):
                backend_aliases.add(alias)
    return [get_scheduler_backend_class(alias)() for alias in backend_aliases]



from django.test import TestCase
from django.utils import timezone
from . import schedules
from datetime import datetime


now = timezone.make_aware(
    datetime(1983, 7, 1, 3, 41),
    timezone.utc)


class ScheduleTest(TestCase):
    def test_hourly(self):
        sched = schedules.Hourly(20, 2, 4)
        self.assertEqual(sched.time_before(now),
                         timezone.make_aware(
                             datetime(1983, 7, 1, 3, 20, 2, 4),
                             timezone.utc))
        self.assertEqual(sched.time_after(now),
                         timezone.make_aware(
                             datetime(1983, 7, 1, 4, 20, 2, 4),
                             timezone.utc))

    def test_every(self):
        sched = schedules.Every(minutes=1)
        self.assertEqual(sched.time_before(now),
                         timezone.make_aware(
                             datetime(1983, 7, 1, 3, 41),
                             timezone.utc))
        self.assertEqual(sched.time_after(now),
                         timezone.make_aware(
                             datetime(1983, 7, 1, 3, 42),
                             timezone.utc))



from django.conf import settings as project_settings
import datetime


_settings = getattr(project_settings, 'PERIODICALLY', {})
DEFAULT_TIMEOUT = _settings.get('DEFAULT_TIMEOUT', datetime.timedelta(hours=1))
SCHEDULERS = _settings.get('SCHEDULERS', {})

# Add the default scheduler it hasn't been explicitly overridden. This way, you
# won't have to redefine the default when you really just want to add a new
# scheduler.
if 'default' not in SCHEDULERS:
    SCHEDULERS['default'] = {'backend': 'periodically.backends.DefaultBackend'}



from django.db import models


class ExecutionRecordManager(models.Manager):
    def get_most_recent(self, task=None, schedule=None):
        qs = self.get_query_set().order_by('-start_time').all()
        if task:
            qs = qs.filter(task_id=task.task_id)
        if schedule:
            qs = qs.filter(schedule_id=schedule.schedule_id)
        return qs[0] if qs else None


class ExecutionRecord(models.Model):
    task_id = models.CharField(max_length=255)
    schedule_id = models.CharField(max_length=32)
    scheduled_time = models.DateTimeField()
    start_time = models.DateTimeField()
    end_time = models.DateTimeField(blank=True, null=True)
    completed_successfully = models.BooleanField(default=False)
    is_fake = models.BooleanField(default=False)

    objects = ExecutionRecordManager()

    def __unicode__(self):
        return '%s @ %s' % (self.task_id, self.start_time)



from __future__ import print_function
from .models import ExecutionRecord
from django.utils import timezone
from .signals import task_complete
from .utils import get_scheduled_time
import logging
import sys


# FIXME: We really need to clarify our use of the word "scheduled." Does it mean scheduled with the backend (as in schedule_task) or scheduled to run (as in run_scheduled_tasks)

class BaseBackend(object):
    """
    Keeps a schedule of periodic tasks.

    """
    _schedules = []

    @property
    def logger(self):
        return logging.getLogger('periodically') # TODO: Further namespace logger?

    @property
    def tasks(self):
        """A list of the tasks scheduled with this backend."""
        return set([task for task, schedule in self._schedules])

    def schedule_task(self, task, schedule):
        """
        Schedules a periodic task.

        """
        # Don't add the same one twice.
        for t, s in self._schedules:
            if t.task_id == task.task_id and s == schedule:
                return

        task_id = task.task_id
        self.logger.info('Scheduling task %s to run on schedule %s' % (task_id,
                schedule))
        self._schedules.append((task, schedule))

        # Subscribe to the task_complete signal. We do this when the task
        # is scheduled (instead of when it runs) so that if you kill Django
        # with unfinished tasks still running, they will be able to
        # complete. (For example, whether a task is completed might be
        # determined by polling a web service. When Django restarts, the
        # polling could start again and the task would be completed.)
        if not getattr(task, 'is_blocking', True):
            task_complete.connect(self._create_receiver(task.__class__),
                    sender=task.__class__, dispatch_uid=task_id)

    def run_scheduled_tasks(self, tasks=None, fake=None):
        """
        Runs any scheduled periodic tasks and ends any tasks that have exceeded
        their timeout. The optional <code>tasks</code> argument allows you to
        run only a subset of the registered tasks.
        """
        self._run_tasks(tasks, fake, False)

    def run_tasks(self, tasks=None, fake=None):
        """
        Run tasks regardless of whether they are scheduled.
        """
        self._run_tasks(tasks, fake, True)

    def _run_tasks(self, tasks=None, fake=None, force=False):
        now = timezone.now()

        # Verify that the provided tasks actually exist.
        if tasks:
            for task in tasks:
                if task.task_id not in set([task.task_id for task in self.tasks]):
                    raise Exception('%s is not registered with this'
                            ' backend.' % task)

        registered_task_ids = [task.task_id for task in tasks]

        for task, schedule in self._schedules:
            if not tasks or task.task_id in registered_task_ids:

                # Cancel the task if it's timed out.
                # FIXME: This should only be called once per task (no matter how many times it's scheduled).
                self.check_timeout(task, now)

                # If there are still tasks running, don't run the queue (as we
                # could mess up the order).
                if ExecutionRecord.objects.filter(end_time__isnull=True):
                    print('There are still tasks running; no new tasks will be run')
                    # TODO: Should this behave differently if force == True?
                    return

                # Run the task if it's due (or past due).
                if force or get_scheduled_time(task, schedule, now) <= now:
                    previous_record = ExecutionRecord.objects.get_most_recent(task, schedule)
                    fake_task = fake or fake is None and previous_record is None

                    # If we're forcing the task, use the previous scheduled
                    # time.
                    scheduled_time = previous_record.scheduled_time if force \
                            and previous_record else None

                    if fake_task:
                        self.fake_task(task, schedule, scheduled_time, now)
                    else:
                        self.run_task(task, schedule, scheduled_time, now)

    def check_timeout(self, task, now=None):
        from .settings import DEFAULT_TIMEOUT

        if now is None:
            now = timezone.now()

        for record in ExecutionRecord.objects.filter(task_id=task.task_id,
                end_time__isnull=True):
            timeout = getattr(task, 'timeout', DEFAULT_TIMEOUT)
            running_time = now - record.start_time
            if running_time > timeout:
                extra = {
                    'level': logging.ERROR,
                    'msg': 'Task timed out after %s.' % running_time,
                }
                self.complete_task(task, False, extra=extra)

    def check_timeouts(self, now=None):
        """
        Checks to see whether any scheduled tasks have timed out and handles
        those that have.
        """
        for task in self.tasks:
            self.check_timeout(task, now)

    def fake_task(self, task, schedule, scheduled_time=None, now=None):
        # TODO: Do we need both of these?
        print('Faking periodic task "%s"' % task.task_id)
        self.logger.info('Faking periodic task "%s"' % task.task_id)

        if scheduled_time is None:
            scheduled_time = get_scheduled_time(task, schedule, now)

        # Create the log for this execution.
        log = ExecutionRecord.objects.create(
            task_id=task.task_id,
            schedule_id=schedule.schedule_id,
            scheduled_time=scheduled_time,
            start_time=now,
            end_time=now,
            is_fake=True,
        )

    def run_task(self, task, schedule, scheduled_time=None, now=None):
        """
        Runs the provided task. This method is provided as a convenience to
        subclasses so that they do not have to implement all of the extra stuff
        that goes with running a single task--for example, retries, failure
        emails, etc. If you want, your subclass's run_scheduled_tasks method
        can call task.run() directly (avoiding this method), but it is highly
        discouraged.

        """
        # TODO: Do we need both of these?
        print('Running periodic task "%s"' % task.task_id)
        self.logger.info('Running periodic task "%s"' % task.task_id)

        if scheduled_time is None:
            scheduled_time = get_scheduled_time(task, schedule, now)

        # Create the log for this execution.
        log = ExecutionRecord.objects.create(
            task_id=task.task_id,
            schedule_id=schedule.schedule_id,
            scheduled_time=scheduled_time,
            start_time=now,
            end_time=None,
        )

        # Run the task.
        try:
            task.run()
        except Exception as err:
            extra = {
                'level': logging.ERROR,
                'msg': str(err),
                'exc_info': sys.exc_info(),
            }
            success = False
        except (KeyboardInterrupt, SystemExit) as err:
            extra = {
                'level': logging.DEBUG,
                'msg': 'The task was cancelled by the user.',
            }
            success = False
            raise
        else:
            extra = None
            success = True
        finally:
            if not success or getattr(task, 'is_blocking', True):
                self.complete_task(task, success, extra=extra)

    def _create_receiver(self, sender):
        def receiver(task, extra=None):
            task_complete.disconnect(receiver, sender,
                    dispatch_uid=task.task_id)
            self.complete_task(task, True, extra=extra)
        return receiver

    def complete_task(self, task, success=True, extra=None):
        """
        Marks a task as complete and performs other post-completion tasks. The
        <code>extra</code> argument is a dictionary of values to be passed to
        <code>Logger.log()</code> as keyword args.

        """
        if extra is not None:
            self.logger.log(**extra)

        record = ExecutionRecord.objects \
                .filter(task_id=task.task_id, end_time=None) \
                .order_by('-start_time')[0]
        record.end_time = timezone.now()
        record.completed_successfully = success
        record.save()

        # TODO: Retries.

class DefaultBackend(BaseBackend):
    """
    A backend that only runs tasks when explicitly told to (i.e. when its
    `run_scheduled_tasks()` method is invoked).

    """
    pass



class PeriodicTask(object):
    backend = 'default'
    is_blocking = True
    
    @property
    def task_id(self):
        return '%s.%s' % (self.__module__, self.__name__)

    def run(self):
        raise RuntimeError('This method must be overridden by your class.')



"""
A collection of functions for quickly registering periodic tasks using the
decorator syntax. Note that, while these functions are used as decorators, they
actually return the original function, so you can decorate multiple times
without worry.
"""

from . import register
from .schedules import Every, Hourly, Daily, Weekly


def _create_decorator(schedule_class, backend=None, *args, **kwargs):
    """
    Creates a decorator that registers periodic tasks. Decorators in this
    module delegate to this method in the interest of DRY: it handles all of
    the common arguments.
    """
    schedule = schedule_class(*args, **kwargs)
    def decorator(fn):
        register.simple_task(fn, schedule, backend=backend)
        return fn
    return decorator


def every(*args, **kwargs):
    """
    A decorator that allows you to specify an arbitrary repeat interval.
    """
    return _create_decorator(Every, *args, **kwargs)


def hourly(*args, **kwargs):
    return _create_decorator(Hourly, *args, **kwargs)


def daily(*args, **kwargs):
    return _create_decorator(Daily, *args, **kwargs)


def weekly(*args, **kwargs):
    return _create_decorator(Weekly, *args, **kwargs)



from __future__ import print_function
from periodically.decorators import *
from periodically import register
from datetime import timedelta


@every(minutes=1)
def task2():
    print('RUNNING exampleapp.periodictasks.task2')
    raise Exception('Custom exception!')


def task3():
    print('RUNNING exampleapp.periodictasks.task3')
register.simple_task(task3, Every(timedelta(weeks=2)))


@daily(hour=19, minute=38)
@daily(hour=19, minute=38) # Duplicate. Shouldn't reschedule.
@daily(hour=20, minute=13) # Not a duplicate. Should schedule again.
def task4():
    print('RUNNING exampleapp.periodictasks.task4')



import inspect
from .tasks import PeriodicTask
from .utils import get_scheduler_backend_class


# Based on django's admin app's autodiscover.
def autodiscover():
    """
    Auto-discover INSTALLED_APPS periodictasks.py modules and fail silently
    when not present.
    """

    from django.conf import settings as project_settings
    from django.utils.importlib import import_module
    from django.utils.module_loading import module_has_submodule

    for app in project_settings.INSTALLED_APPS:
        mod = import_module(app)
        # Attempt to import the app's periodictasks module.
        try:
            import_module('%s.periodictasks' % app)
        except:
            # Decide whether to bubble up this error. If the app just
            # doesn't have a periodictasks module, we can ignore the error
            # attempting to import it, otherwise we want it to bubble up.
            if module_has_submodule(mod, 'periodictasks'):
                raise


class TaskRegistry(object):

    _backend_singletons = {} # Maps backend classes to instances.
    _backends = set()

    @property
    def backends(self):
        return self._backends

    def simple_task(self, fn, schedule, backend=None):
        """
        Registers a callable as a periodic task.
        """
        _task_id = '%s.%s' % (fn.__module__, fn.__name__)

        # Create a PeriodicTask subclass that wraps the function.
        class DecoratedTask(PeriodicTask):
            task_id = _task_id

            def run(self, *args, **kwargs):
                fn(*args, **kwargs)

        task_instance = DecoratedTask()
        self.task(task_instance, schedule, backend)
        return task_instance

    def task(self, task, schedule, backend=None):
        """
        Registers a periodic task.
        """
        if inspect.isclass(task):
            task = task()
            
        backend_class = get_scheduler_backend_class(backend)
        try:
            backend = self._backend_singletons[backend_class]
        except KeyError:
            backend = self._backend_singletons[backend_class] = backend_class()
        self._backends.add(backend)
        backend.schedule_task(task, schedule)


register = TaskRegistry()
autodiscover()



#!/usr/bin/env python

import os
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

README = read('README.rst')

setup(
    name = "django-periodically",
    version = "0.3.0",
    description='Periodic task management for your Django projects.',
    url = 'https://github.com/hzdg/django-periodically',
    long_description=README,

    author = 'Matthew Tretter',
    author_email = 'matthew@exanimo.com',
    packages = find_packages()
)



import time as _time
from django.utils import timezone
from datetime import datetime, date, time, timedelta
from hashlib import md5


class Schedule(object):
    def __eq__(self, other):
        other_id = getattr(other, 'schedule_id', None)
        return self.schedule_id == other_id

    def __ne__(self, other):
        return not self.__eq__(other)


class BaseSchedule(object):
    """A base class for the built-in schedules. Don't use this yourself.

    """

    @property
    def schedule_id(self):
        class_name = '%s.%s' % (self.__class__.__module__,
                self.__class__.__name__)
        time_args = tuple([getattr(self, name) for name in self._time_attrs])
        return md5(str((class_name, time_args))).hexdigest()

    def time_before(self, time):
        kwargs = dict((k, getattr(self, k)) for k in self._time_attrs)
        t = time.replace(**kwargs)
        if t > time:
            previous_time = t - self.repeat_interval
        else:
            previous_time = t
        return previous_time

    def time_after(self, time):
        return self.time_before(time) + self.repeat_interval


class Hourly(BaseSchedule):
    repeat_interval = timedelta(hours=1)
    _time_attrs = ['minute', 'second', 'microsecond']

    def __init__(self, minute=0, second=0, microsecond=0):
        super(Hourly, self).__init__()
        self.minute = minute
        self.second = second
        self.microsecond = microsecond

    def __str__(self):
        return 'Hourly @ %s' % time(minute=self.minute, second=self.second,
                microsecond=self.microsecond)


class Daily(BaseSchedule):
    repeat_interval = timedelta(days=1)
    _time_attrs = ['hour', 'minute', 'second', 'microsecond']

    def __init__(self, hour=0, minute=0, second=0, microsecond=0):
        super(Daily, self).__init__()
        self.hour = hour
        self.minute = minute
        self.second = second
        self.microsecond = microsecond

    def __str__(self):
        return 'Daily @ %s' % time(hour=self.hour, minute=self.minute,
                second=self.second, microsecond=self.microsecond)


class Weekly(BaseSchedule):
    repeat_interval = timedelta(days=7)
    _time_attrs = ['day', 'hour', 'minute', 'second', 'microsecond']

    def __init__(self, day=0, hour=0, minute=0, second=0, microsecond=0):
        super(Weekly, self).__init__()
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second
        self.microsecond = microsecond

    def __str__(self):
        day_name = _time.strftime('%A', _time.strptime(str(self.day), '%w'))
        return '%s @ %s' % (day_name, time(hour=self.hour, minute=self.minute,
                second=self.second, microsecond=self.microsecond))


def total_seconds(td):
    """Since ``timedelta.total_seconds()`` is new in 2.7"""
    return float(td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**6


class Every(BaseSchedule):
    _time_attrs = ['repeat_interval']

    def __init__(self, interval=None, days=0, seconds=0, microseconds=0,
            milliseconds=0, minutes=0, hours=0, weeks=0,
                 starting_at=timezone.make_aware(datetime(1970, 1, 1),
                                                 timezone.utc)):
        super(Every, self).__init__()
        if interval is None:
            interval = timedelta(days, seconds, microseconds,
                milliseconds, minutes, hours, weeks)
        self.repeat_interval = interval
        self.starting_at = starting_at

    def time_before(self, time):
        # Uses "perfect time" (no leap years, etc.)
        interval = self.repeat_interval.microseconds / 1000 \
                + total_seconds(self.repeat_interval) * 1000
        time_since_start = time - self.starting_at
        ms_since_start = time_since_start.microseconds / 1000 \
                + total_seconds(time_since_start) * 1000
        delta_in_ms = interval * int(ms_since_start / interval)
        return self.starting_at + timedelta(milliseconds=delta_in_ms)

    def __str__(self):
        return 'Every %s' % self.repeat_interval



from django.core.management.base import BaseCommand, CommandError
from ... import register as task_scheduler
from optparse import make_option
from ...utils import get_scheduler_backends_in_groups


class InvalidBackendGroupError(Exception):
    def __init__(self, backend_group):
        Exception.__init__(self, '%s is not a valid backend group' % backend_group)


class InvalidBackendError(Exception):
    def __init__(self, backend):
        Exception.__init__(self, 'Could not find backend %s' % backend)


class Command(BaseCommand):
    help = "Runs scheduled tasks."

    option_list = BaseCommand.option_list + (
        make_option('--group',
            dest='backend_groups',
            type='string',
            action='append',
            default=None,
            help='The scheduler group that should run its scheduled tasks. Groups can be defined in your settings.py file.',
        ),
        make_option('--force',
            dest='force_execution',
            action='store_true',
            default=False,
            help='Use this flag to run tasks regardless of whether they are scheduled to be run.',
        ),
        make_option('--fake',
            dest='fake',
            action='store_true',
            default=None,
            help='Use this flag to create execution records for tasks without actually running them. If the flag is not used, the backend will decide whether to fake tasks.',
        ),
    )

    def handle(self, *args, **options):
        task_ids = args
        backend_groups = options.get('backend_groups', None)
        fake = options['fake']
        force_execution = options['force_execution']
        
        if backend_groups:
            backends = get_scheduler_backends_in_groups(backend_groups)
        else:
            backends = task_scheduler.backends
        
        for backend in backends:
            if task_ids:
                tasks = set([task for task in backend.tasks if task.task_id in task_ids])
            else:
                tasks = backend.tasks
            if force_execution:
                backend.run_tasks(tasks, fake=fake)
            else:
                backend.run_scheduled_tasks(tasks, fake=fake)
